from population import Population
from individual import Individual
import matplotlib.pyplot as plt
import time
import numpy as np
from metrics import metrics
from tools import *
import copy

class MOEAD:

    def __init__(self,problem, popsize, max_evo):
        self.PL = [] # current population
        self.PP = [] # pareto front to be optimized
        self.PE = [] # external archive
        self.refpoint = []
        self.populations = Population(popsize,problem)
        self.idv = Individual()
        self.popsize = popsize
        self.max_evo = max_evo
        self.problem = problem
        self.metric = metrics('MOEAD')
        # weight
        #self.weights = Load_W(self.popsize, 2) # objective is 2
        self.weights = generate_W(self.popsize)
        self.T = 5
        self.neighbor = self.cpt_W_Bi_T(self.weights, self.T)


    # T neighbors
    def cpt_W_Bi_T(self, W, T):
        W_Bi_T = []
        for bi in range(W.shape[0]):
            Bi = W[bi]
            DIS = np.sum((W - Bi) ** 2, axis=1)
            B_T = np.argsort(DIS)
            B_T = B_T[1:T + 1]
            W_Bi_T.append(B_T)
        return W_Bi_T

    def is_dominate(self, a, b):
        a_f = a.F_value
        b_f = b.F_value
        i = 0
        for av, bv in zip(a_f, b_f):
            if av < bv:
                i = i + 1
            if av > bv:
                return False
        if i != 0:
            return True
        return False
    
    def find_pf(self, P):
        F1 = []
        for j, p in enumerate(P):
            n = 0
            for i, q in enumerate(P):
                if j != i:
                    if self.is_dominate(q, p):
                        n += 1
            if n == 0:
                F1.append(p)
        return F1       

    def update_pf(self, pop, cp):
        pop_value = [p.F_value for p in pop]
        if cp.F_value not in pop_value:
            pop.append(cp)
            new_pop = self.find_pf(pop)
        else:
            new_pop = pop
        return new_pop

    def is_better(self, i, p1, p2):
        qbxf1 = 0
        qbxf2 = 0
        for m in range(2):
            qbxf1 +=  normalization(p1.F_value[m], self.populations.nadirpoint[m], self.populations.refpoint[m]) * self.weights[i][m]
            qbxf2 +=  normalization(p2.F_value[m], self.populations.nadirpoint[m], self.populations.refpoint[m]) * self.weights[i][m]
        if qbxf1 < qbxf2:
            return True
        else:
            return False

    def impove_solution(self, i, p):
        cp = self.idv.copy_one(p)
        s_truck = cp.X[0]
        s_drone = cp.X[1]        
        operation = random.choice(['relocation','2opt','id_greedy','id_random'])
        [new_truck, new_drone, s] = self.problem.neighborhoods(s_truck, s_drone, operation)
        cp.X[0] = new_truck
        cp.X[1] = new_drone
        cp.F_value = s
        if self.is_better(i, cp, p):
            return cp
        else:
            return p

    def run(self):
        # start time
        start_time = time.time()
        # initialization
        self.Pop = self.populations.creat_pop(self.popsize, refpoint=True, weights = self.weights)
        self.PE = self.find_pf(self.Pop)
        # iterations
        gen = 0
        print('%s th pareto len %s:' % (gen, len(self.PE)))
        while gen < self.max_evo:
            gen += 1
            for pi, p in enumerate(self.Pop):
                Bi = self.neighbor[pi]
                k = np.random.randint(self.T)
                l = np.random.randint(self.T)        
                ik = Bi[k]
                il = Bi[l]
                # Xi = self.Pop[pi]
                Xk = self.Pop[ik]
                Xl = self.Pop[il]
                Y = self.populations.reproduction(Xk, Xl)
                
                if len(Y) == 1:
                    y = Y[0]
                else:
                    y1 = Y[0]
                    y2 = Y[1]
                    if self.is_better(pi, y1, y2):
                        y = y1
                    else:
                        y = y2

                # # improve solution
                # if gen % 5 == 0:
                #     y = self.impove_solution(pi,y)
                # update nadinode point
                if y.F_value[0] > self.populations.nadirpoint[0]:
                    self.populations.nadirpoint[0] = y.F_value[0]
                if y.F_value[1] > self.populations.nadirpoint[1]:
                    self.populations.nadirpoint[1] = y.F_value[1]
                if y.F_value[0] < self.populations.refpoint[0]:
                    self.populations.refpoint[0] = y.F_value[0]
                if y.F_value[1] < self.populations.refpoint[1]:
                    self.populations.refpoint[1] = y.F_value[1]
                
                # update neighborhood solutions
                for j in Bi:
                    if self.is_better(j, y, self.Pop[j]):
                        self.Pop[j] = y
                # update EP
                self.PE = self.update_pf(self.PE, y)
                
            print('%s th pareto len %s:' % (gen, len(self.PE)))
            # record
            self.metric.record(self.PE, gen)
            # update reference point
            #self.populations.update_refpoint(self.PL)

        end_time = time.time()
        used_time = end_time - start_time
        print('used time:', used_time)
        self.metric.recordTime(used_time)
        self.metric.recordfinal(self.PE)


    def draw(self,problemname, outpath, repeat):

        pf1_data = []
        pf2_data = []

        for p in self.PE:
            pf1_data.append(p.F_value[0])
            if p.F_value[1] != 0:
                pf2_data.append(p.F_value[1])
            else:
                pf2_data.append(0)
        f1_data = []
        f2_data = []
        for a in self.PL:
            f1_data.append(a.F_value[0])
            if a.F_value[1] != 0:
                f2_data.append(a.F_value[1])
            else:
                f2_data.append(0)
        
        plt.figure() 
        plt.xlabel('Cost', fontsize=15)
        plt.ylabel('-Customer satisfaction', fontsize=15)

        plt.scatter(f1_data, f2_data, c='black', s=5)
        plt.scatter(pf1_data, pf2_data, c='red', s=10)
        plt.tight_layout()
        #plt.show()
        #plt.savefig(outpath + problemname.split('.')[0] + '_' + str(repeat) + '.jpg', dpi = 300)


    def plot_map(self):
        color_ = ['#FF0000','#BC8F8F','#FA8072','#FAA460','#2E8B57','#FFF5EE','#A0522D','#C0C0C0','#87CEEB','#6A5ACD','#708090']
        # map
        X = []
        Y = []
        for cus in self.problem.data_dict:
            x = self.problem.data_dict[cus]['latDeg']
            y = self.problem.data_dict[cus]['lonDeg']
            X.append(x)
            Y.append(y)

        p = self.PE[int(len(self.PE)/2)]
        print('chorm_vehicle:', p.X[0])
        print('chorm_drone:', '\n',p.X[1])
        if not self.problem.check_feasible(p.X[0], p.X[1]):
            print('error occurs')
            

        s_truck = p.X[0]
        s_drone = p.X[1]

        s_truck = np.append(s_truck, s_truck[0])

        X_truck = []
        Y_truck = []

        for cus in s_truck:
            x = self.problem.data_dict[cus]['latDeg']
            y = self.problem.data_dict[cus]['lonDeg']
            X_truck.append(x)
            Y_truck.append(y)

        plt.figure()
        plt.scatter(X,Y)
        plt.plot(X_truck,Y_truck,color = 'blue',label = 'Truck route')

        drawed = []
        
        for index_ in range(s_drone.shape[1]):
            i_index = s_drone[0][index_]      
            k_index = s_drone[2][index_]
            i       = s_truck[i_index]
            j       = s_drone[1][index_]   
            k       = s_truck[k_index]
            d       = s_drone[3][index_]

            X2 = []
            Y2 = []

            ix = self.problem.data_dict[i]['latDeg']
            iy = self.problem.data_dict[i]['lonDeg']
            X2.append(ix)
            Y2.append(iy)
            jx = self.problem.data_dict[j]['latDeg']
            jy = self.problem.data_dict[j]['lonDeg']
            X2.append(jx)
            Y2.append(jy)
            kx = self.problem.data_dict[k]['latDeg']
            ky = self.problem.data_dict[k]['lonDeg']
            X2.append(kx)
            Y2.append(ky)

            if index_ != s_drone.shape[1] - 1:
                if d not in drawed:
                    plt.plot(X2,Y2,'--',color = color_[d], label = 'Drone {} route'.format(d+1))
                    drawed.append(d)
                else:
                    plt.plot(X2,Y2,'--',color = color_[d])

        # last drone route
        if d not in drawed:
            plt.plot(X2,Y2,'--',color = color_[d], label = 'Drone {} route'.format(d+1))
        else:
            plt.plot(X2,Y2,'--',color = color_[d])

        # depot = 0
        depot_x = [self.problem.data_dict[0]['latDeg']]
        depot_y = [self.problem.data_dict[0]['lonDeg']]
        plt.scatter(depot_x, depot_y, c='yellow', marker='*')
        plt.legend()
        plt.tight_layout()

        plt.show()


    def save(self, time, problemname, outpath):
        pf1_data = []
        pf2_data = []

        for p in self.PE:
            pf1_data.append(str(p.F_value[0]))
            pf2_data.append(str(p.F_value[1]))

        with open(outpath + problemname.split('.')[0] + '_results.csv', 'a+') as f:
            data = ','.join(pf1_data)
            f.write(data)
            f.write('\t')
            data = ','.join(pf2_data)
            f.write(data)
            f.write('\t')
            f.write(str(time))
            f.write('\n')
        #with open(outpath + problemname.split('.')[0] + '_RunTime.csv', 'a+') as f:

    
    def draw_HV(self):
        # paretoFs = self.metric.paretoFs
        # refpoint = self.metric.define_refpoints(paretoFs)
        worest_node = np.array([999, 0])
        best_node = np.array([0, -(len(self.problem.data_dict) - 1)])
        self.metric.cal_HV(worest_node, best_node)

        X = self.metric.gen
        Y = self.metric.HVs

        #plt.figure()
        plt.plot(X,Y, label = 'HV')
        plt.xlabel('Generations', fontsize=15)
        plt.ylabel('Hypervolume', fontsize=15)
        plt.legend()
        plt.tight_layout()
        plt.show()

    def draw_cost(self):
        self.metric.cal_fitness()

        X = self.metric.gen
        Y1 = self.metric.MeanCost
        Y2 = self.metric.MinCost

        plt.figure()
        plt.plot(X,Y1, label = 'MeanCost')
        plt.xlabel('Generations', fontsize=15)
        plt.ylabel('MeanCost', fontsize=15)
        plt.legend()
        plt.tight_layout()
        plt.show()

        plt.figure()
        plt.plot(X,Y2, label = 'MinCost')
        plt.xlabel('Generations', fontsize=15)
        plt.ylabel('MinCost', fontsize=15)
        plt.legend()
        plt.tight_layout()
        plt.show()