import numpy as np
from tools import *
import copy
from collections import Counter, defaultdict
from individual import Individual

class MyProblem():
    def __init__(self, problemName,file_path):
        # parameters settings
        self.wbl = 0.2
        self.wbu = 0.2
        self.drone_cost = 1/25
        self.truck_cost = 1
        self.drone_num = 3 # drone numbers
        self.truck_service_time = 10
        self.idv = Individual()
        # read data
        [data_dict, droneMatrix, truckMatrix] = read_data(problemName, file_path, self.wbl, self.wbu)
        self.data_dict = data_dict
        self.droneMatrix = droneMatrix
        self.truckMatrix = truckMatrix
        
        # truck_only customers
        self.truck_only_cus = []
        for cus_num in self.data_dict:
            drone_enable = self.data_dict[cus_num]['drone_enable']
            if drone_enable == 0:
                # depot or truck-only-node
                self.truck_only_cus.append(cus_num)

        # lowest number of truck customers
        self.LTL = np.ceil((len(self.data_dict)-self.drone_num)/(self.drone_num + 1))
        #calculate cost matrix and flight range
        self.node_num = len(self.data_dict)
        self.drone_endurance = 10 # initial value, not final value
        self.feasible_flight_rate = 0.35
        self.m_cost = np.zeros((self.node_num, self.node_num, self.node_num))
        while True:
            for i in range(len(self.data_dict)):
                for j in range(len(self.data_dict)):
                    for k in range(len(self.data_dict)):
                        if i != k and i != j and k != j:
                            if self.droneMatrix[i][j] + self.droneMatrix[j][k] <= self.drone_endurance:
                                if j not in self.truck_only_cus:
                                    self.m_cost[i][j][k] = self.drone_cost * (self.droneMatrix[i][j] + self.droneMatrix[j][k])
                                else:
                                    self.m_cost[i][j][k] = float('inf')
                            else:
                                self.m_cost[i][j][k] = float('inf')
                        else:
                            self.m_cost[i][j][k] = float('inf')
            feasible_shots = np.sum(self.m_cost != float('inf'))
            if feasible_shots/(self.node_num ** 3) < self.feasible_flight_rate:
                self.drone_endurance += 1
            else:
                break
    
###### generate initial chromosomes ########
    def initialsolution(self):
        while True:
            [truck_cus, drone_cus] = self.divide_customer()
            [s_truck, s_drone] = self.assign_drones(truck_cus,drone_cus)
            if len(list(s_drone)) != 0:
                break
        s_best = self.evaluate(s_truck, s_drone)
        s_chromosome = [s_truck, s_drone]
        return [s_best, s_chromosome]

    def divide_customer(self):
        drone_cus = []
        truck_cus = []

        # divide customers
        for cus_num in self.data_dict:
            if cus_num not in self.truck_only_cus:
                drone_cus.append(cus_num)

        truck_cus = truck_cus + self.truck_only_cus

        while len(truck_cus) < self.LTL:
            random_cus = random.choice(drone_cus)
            truck_cus.append(random_cus)
            drone_cus.remove(random_cus)

        truck_cus = nearest_neighbor_tw(truck_cus, self.truckMatrix, self.data_dict)

        return [truck_cus, drone_cus]

    def assign_drones(self, truck_cus, drone_cus):
        # greedy assign drone shots
        # for each drone, assign cheapest shot to it.
        feasible = 0

        while feasible == 0:
            nodes = truck_cus + [0]
            P = []
            drone_cus_copy = copy.deepcopy(drone_cus)

            for d in range(self.drone_num):
                i_index = 0
                while True:
                    possible_shot = []
                    for k_index in range(i_index + 1, len(nodes)):
                        i = nodes[i_index]
                        k = nodes[k_index]
                        for j in drone_cus_copy:
                            cost = self.m_cost[i][j][k]
                            if cost != float('inf'):
                                possible_shot.append([cost, i_index, j, k_index, d])
                            
                    if possible_shot != []:
                        P_d = sorted(possible_shot, key = lambda x: x[0]) # find the cheapest shot
                        shot = P_d[0]
                        j = shot[2]
                        k_index = shot[3]
                        drone_cus_copy.remove(j)
                        i_index = k_index
                        P.append(shot[1:])

                    if drone_cus_copy == [] or k_index == len(nodes) - 1:
                        break
            
            if drone_cus_copy == []:
                # generate a solution           
                s_drone = np.array(P).T
                s_truck = np.array(truck_cus)
                feasible = 1
            else:
                truck_cus = list(truck_cus) + drone_cus_copy
                drone_cus = list(set(drone_cus) - set(drone_cus_copy))
                truck_cus = nearest_neighbor_tw(truck_cus,self.truckMatrix, self.data_dict)

        s_truck = np.append(s_truck, 0)
        return [s_truck, s_drone]                       

##### evaluate chromosome #######
    def evaluate(self, s_truck, s_drone):
        truck_leave_time = {}
        t = 0
        cost = 0
        satisfaction = 0
        
        # leave time of depot is 0
        truck_leave_time[0] = t
        
        # truck cost
        for i_index in range(len(s_truck) - 1):
            i = s_truck[i_index]
            k = s_truck[i_index + 1]
            truck_service_time = self.truck_service_time
            truck_earliest_time = self.data_dict[k]['earliest_time']
            cost += self.truck_cost * self.truckMatrix[i][k]
            t_truck = t + self.truckMatrix[i][k]
            if k != 0:
                satisfaction += evaluate_satisfaction(self.data_dict, k, t_truck)

            rendezvous_index = np.argwhere(s_drone[2] == i_index + 1) 
            if list(rendezvous_index) != []: # if drones land at node k
                drone_reach_time = []
                for x in rendezvous_index:
                    x = int(x)
                    j  = int(s_drone[1][x])
                    i2 = int(s_truck[s_drone[0][x]])
                    cost += self.m_cost[i2][j][k]
                    t_drone = truck_leave_time[i2] + self.droneMatrix[i2][j]
                    satisfaction += evaluate_satisfaction(self.data_dict, j, t_drone)

                    if t_drone <= self.data_dict[j]['earliest_time']:
                        # drone has to wait
                        wait_time = self.data_dict[j]['earliest_time'] - t_drone
                    else:
                        wait_time = 0

                    j2k = t_drone + wait_time + self.droneMatrix[j][k]
                    drone_reach_time.append(j2k)

            else:
                # no drones land on node k
                drone_reach_time = [0]

            t = max([t_truck + truck_service_time, truck_earliest_time + truck_service_time, max(drone_reach_time)])
            if k != 0:
                truck_leave_time[k] = t
            
        return [cost, -satisfaction]

##### check feasible and repair chromosome #####    

    def not_conflict(self, s_drone, pair):
        # each node should launch less than drone_num drones, same as to rendezvous node
        i_index = pair[0]
        k_index = pair[2]
        d = pair[3]
        position1 = np.argwhere(s_drone[0] == i_index)
        position2 = np.argwhere(s_drone[2] == k_index)
        position3 = np.argwhere(s_drone[3] == d)
        if len(position1) < self.drone_num and len(position2) < self.drone_num:
            for shot_index in position3:
                launch_index = s_drone[0][shot_index]
                land_index   = s_drone[2][shot_index]
                # case 1
                if launch_index <= i_index < land_index:
                    return False
                # case 2
                if launch_index < k_index <= land_index:
                    return False
                # case 3, include
                if launch_index > i_index and land_index < k_index:
                    return False
        else:
            return False
        return True

    def check_feasible(self, s_truck, s_drone):
        # check if a chromosome is feasible or not
        infeasible_index = []
        infeasible_j     = []
        nodes = np.append(s_truck,0)
        for index_ in range(s_drone.shape[1]):
            i = nodes[s_drone[0][index_]]
            j = s_drone[1][index_]
            k = nodes[s_drone[2][index_]]
            if self.m_cost[i][j][k] == float('inf'):
                infeasible_index.append(index_)
                infeasible_j.append(j)
        if infeasible_j == []:
            return [True, None, None]
        else:
            return [False, infeasible_index, infeasible_j]

    def reassign(self, s_truck, s_drone, infeasible_j, infeasible_index = None):

        if infeasible_index != None:
            # infeasible, update s_drone
            s_drone_new = np.delete(s_drone, infeasible_index, axis = 1)
        else:
            # input modified chromosome directly
            s_drone_new = copy.deepcopy(s_drone)
        s_truck_new = copy.deepcopy(s_truck)
        
        # re_assign
        assigned_j = []
        for j in infeasible_j:
            possible_shot = []
            for i_index in range(len(s_truck_new) - 1):
                i = s_truck_new[i_index]
                for k_index in range(i_index + 1, len(s_truck_new)):
                    k = s_truck_new[k_index]
                    cost = self.m_cost[i][j][k]
                    if cost != float('inf'):
                        for d in range(self.drone_num):
                            pair = [i_index, j, k_index, d]
                            if self.not_conflict(s_drone_new, pair):
                                possible_shot.append([cost, i_index,j,k_index, d])

            # greedy assignment
            if possible_shot != []:
                shots_sorted = sorted(possible_shot, key = lambda x: x[0])
                shot = shots_sorted[0]
                s_drone_new = np.column_stack((s_drone_new, shot[1:]))
                j = shot[2]
                assigned_j.append(j)
            else:
                return [False, None, None]

        if len(assigned_j) == len(infeasible_j):
                return [True, s_truck_new, s_drone_new]
        else:
            return [False, None, None]      

    def update_shots(self, pair, shots, s_drone):
        new_shots = defaultdict(list)
        j = pair[1]
        d = pair[3]
        for key in shots:
            if key != j:
                for pair1 in shots[key]:
                    d1 = pair1[4]
                    if d1 != d:
                        new_shots[key].append(pair1)
                    elif self.not_conflict(s_drone, pair1[1:]):
                        new_shots[key].append(pair1)
    
        return new_shots


########### neighborhood operators ############

    def genetic_t2d(self, s_truck, s_drone):
        most_expensive = 0
        removed_index = 0
        for index_ in range(1, len(s_truck)-1):
            j = s_truck[index_]
            if j not in self.truck_only_cus:
                i = s_truck[index_ - 1]
                k = s_truck[index_ + 1]
                cost = self.truckMatrix[i][j] + self.truckMatrix[j][k]
                if cost > most_expensive:
                    most_expensive = cost
                    removed_index  = index_
                    removed_j      = j

        j_index = removed_index
        j = removed_j

        j_to_assign = set()
        # # randomly select a node from truck route
        # available_nodes = list(set(s_truck) - set(self.truck_only_cus))
        # j = random.choice(available_nodes)   
        # j_index = int(np.argwhere(s_truck == j))
        j_to_assign.add(j)

        # updat s_truck
        s_truck_new = np.delete(s_truck, j_index)
        # update s_drone
        index_to_move = set()
        s_drone_new = copy.deepcopy(s_drone)
        for index_ in range(s_drone_new.shape[1]):
            i_index = s_drone_new[0][index_]
            k_index = s_drone_new[2][index_]
            if i_index > j_index:
                s_drone_new[0][index_] -= 1
            elif i_index == j_index:
                index_to_move.add(index_)
                j = s_drone_new[1][index_]
                j_to_assign.add(j)
            if k_index > j_index:
                s_drone_new[2][index_] -= 1
            elif k_index == j_index:
                index_to_move.add(index_)
                j = s_drone_new[1][index_]
                j_to_assign.add(j)

        j_to_assign   = list(j_to_assign)
        index_to_move = list(index_to_move)
        if len(index_to_move) != 0:
            [flag, s_truck_current, s_drone_current] = self.reassign(s_truck_new, s_drone_new, infeasible_j= j_to_assign, infeasible_index= index_to_move)
        else:
            [flag, s_truck_current, s_drone_current] = self.reassign(s_truck_new, s_drone_new, infeasible_j= j_to_assign)
        if flag == True:
            return [s_truck_current, s_drone_current]
        else:
            return [s_truck, s_drone]

    def genetic_d2t(self, s_truck, s_drone):
        most_expensive = 0
        removed_index = 0
        for index_ in range(s_drone.shape[1]):
            i_index = s_drone[0][index_]
            i = s_truck[i_index]
            j = s_drone[1][index_]
            k_index = s_drone[2][index_]
            k = s_truck[k_index]
            cost = self.m_cost[i][j][k]
            if cost > most_expensive:
                most_expensive = cost
                removed_j = j
                removed_index = index_

        j          = removed_j
        node_index = removed_index

        # node_index = random.choice([i for i in range(s_drone.shape[1])])
        # j = s_drone[1][node_index]

        # find potential good position for truck
        cost_save = []
        for i_index in range(1, len(s_truck) - 1):
            i = s_truck[i_index]
            k = s_truck[i_index + 1]
            cost = self.truckMatrix[i][j] + self.truckMatrix[j][k] - self.truckMatrix[i][k]
            cost_save.append([cost, i_index])
        
        cost_save = sorted(cost_save, key = lambda x: x[0])
        insert_index = cost_save[0][1] + 1
        s_truck_new = np.insert(s_truck, insert_index, j)

        # update s_drone
        s_drone_new = np.delete(s_drone, node_index, axis = 1)
        for index_ in range(s_drone_new.shape[1]):
            i_index = s_drone_new[0][index_]
            k_index = s_drone_new[2][index_]
            if i_index >= insert_index:
                s_drone_new[0][index_] += 1
            if k_index >= insert_index:
                s_drone_new[2][index_] += 1               

        return [s_truck_new, s_drone_new]

    def genetic_swap(self,s_truck, s_drone):      
        a_index = random.choice([i for i in range(s_drone.shape[1])])
        while True:
            b_index = random.choice([i for i in range(1,len(s_truck))])
            if s_truck[b_index] not in self.truck_only_cus:
                break
        
        s_truck_new = copy.deepcopy(s_truck)
        s_drone_new = copy.deepcopy(s_drone)

        # update
        s_drone_new[1][a_index], s_truck_new[b_index] = s_truck_new[b_index], s_drone_new[1][a_index]
        [flag1, infeasible_index, infeasible_j] = self.check_feasible(s_truck_new, s_drone_new)
        if flag1 == False:
            [flag2, s_truck_re, s_drone_re] = self.reassign(s_truck_new, s_drone_new, infeasible_j = infeasible_j, infeasible_index= infeasible_index)
            if flag2 == True:
                return [s_truck_re, s_drone_re]
            else:
                return [s_truck, s_drone]
        else:
            return [s_truck_new, s_drone_new]

    # def n_exchange(self, s_truck, s_drone):
    #     nodes = np.append(s_truck, s_drone[1])
    #     len_truck = len(s_truck)
    #     # Swap (1-1). Two nodes exchange their positions.
    #     a_index = random.choice([i for i in range(1, len(nodes))])
    #     while True:
    #         b_index = random.choice([i for i in range(1, len(nodes))])
    #         if a_index != b_index:
    #             break
    #     nodes[a_index], nodes[b_index] = nodes[b_index], nodes[a_index]

    #     # update
    #     s_truck_new = nodes[:len_truck]
    #     s_drone_new = copy.deepcopy(s_drone)
    #     s_drone_new[1] = nodes[len_truck:]

    #     # # check and repair
    #     # [flag, s_truck_re, s_drone_re] = self.check_and_reassign(s_truck_new, s_drone_new)
    #     # if flag == True:
    #     #     return [s_truck_re, s_drone_re]
    #     # else:
    #     #     return [s_truck, s_drone]
    #     return [s_truck_new, s_drone_new] 

        # # check and repair
        # [flag1, infeasible_index, infeasible_j] = self.check_feasible(s_truck_new, s_drone_new)
        # if flag1 == False:
        #     [flag2, s_truck_re, s_drone_re] = self.reassign(s_truck_new, s_drone_new, infeasible_j= infeasible_j, infeasible_index=infeasible_index)
        #     if flag2 == True:
        #         return [s_truck_re, s_drone_re]
        #     else:
        #         return [s_truck, s_drone]
        # else:
        #     return [s_truck_new, s_drone_new]

    def n_relocation_truck_route(self, s_truck, s_drone):
        a_index = random.choice([i for i in range(1,len(s_truck)-1)])
        while True:
            b_index = random.choice([i for i in range(1,len(s_truck)-1)])
            if a_index != b_index:
                break
        # update
        selected_cus = s_truck[b_index]
        s_truck_new = np.delete(s_truck, b_index)
        s_truck_new = np.insert(s_truck_new, a_index, selected_cus)
        s_drone_new = copy.deepcopy(s_drone)

        # check and repair
        [flag1, infeasible_index, infeasible_j] = self.check_feasible(s_truck_new, s_drone_new)
        if flag1 == False:
            [flag2, s_truck_re, s_drone_re] = self.reassign(s_truck_new, s_drone_new, infeasible_j= infeasible_j, infeasible_index=infeasible_index)
            if flag2 == True:
                return [s_truck_re, s_drone_re]
            else:
                return [s_truck, s_drone]
        else:
            return [s_truck_new, s_drone_new]

    def n_2opt(self, s_truck, s_drone):
        truck_cus = list(copy.deepcopy(s_truck))
        truck_cus = two_opt(truck_cus)
        s_truck_new = np.array(truck_cus)
        s_drone_new = copy.deepcopy(s_drone)
        # check and repair
        [flag1, infeasible_index, infeasible_j] = self.check_feasible(s_truck_new, s_drone_new)
        if flag1 == False:
            [flag2, s_truck_re, s_drone_re] = self.reassign(s_truck_new, s_drone_new, infeasible_j= infeasible_j, infeasible_index=infeasible_index)
            if flag2 == True:
                return [s_truck_re, s_drone_re]
            else:
                return [s_truck, s_drone]
        else:
            return [s_truck_new, s_drone_new]

    def n_idgreedy(self, s_truck, s_drone):
        # find the most expensive shot
        most_expensive = 0
        removed_index = 0
        for index_ in range(s_drone.shape[1]):
            i_index = s_drone[0][index_]
            i = s_truck[i_index]
            j = s_drone[1][index_]
            k_index = s_drone[2][index_]
            k = s_truck[k_index]
            cost = self.m_cost[i][j][k]
            if cost > most_expensive:
                most_expensive = cost
                removed_j = j
                removed_index = index_

        # update drone and truck list
        s_drone_new = np.delete(s_drone, removed_index, axis = 1)
        s_truck_new = copy.deepcopy(s_truck)

        # re_assign
        # find all the possible combinations
        possible_shot = []
        j = int(removed_j)
        for i_index in range(len(s_truck_new) - 1):
            i = s_truck_new[i_index]
            for k_index in range(i_index + 1, len(s_truck_new)):
                k = s_truck_new[k_index]
                cost = self.m_cost[i][j][k]
                if cost != float('inf'):
                    for d in range(self.drone_num):
                        pair = [i_index, j, k_index, d]
                        if self.not_conflict(s_drone_new, pair):
                            possible_shot.append([cost, i_index,j,k_index, d])
        # assignment shot combination
        # nodes that have less combinations and less cost are preferred
        pairs = sorted(possible_shot, key = lambda x:x[0])
        pair = pairs[0]
        s_drone_new = np.column_stack((s_drone_new, pair[1:]))
        return [s_truck_new, s_drone_new]    

    def n_idrandom(self, s_truck, s_drone):
        node_index = random.sample(range(s_drone.shape[1]),1)
        removed_j = s_drone[1][node_index]
        # update drone and truck list
        s_drone_new = np.delete(s_drone, node_index, axis = 1)
        s_truck_new = copy.deepcopy(s_truck)

        # re_assign
        # find all the possible combinations
        possible_shot = []
        j = int(removed_j)
        for i_index in range(len(s_truck_new) - 1):
            i = s_truck_new[i_index]
            for k_index in range(i_index + 1, len(s_truck_new)):
                k = s_truck_new[k_index]
                cost = self.m_cost[i][j][k]
                if cost != float('inf'):
                    for d in range(self.drone_num):
                        pair = [i_index, j, k_index, d]
                        if self.not_conflict(s_drone_new, pair):
                            possible_shot.append([cost, i_index,j,k_index, d])
        # assignment shot combination
        pair = random.choice(possible_shot)
        s_drone_new = np.column_stack((s_drone_new, pair[1:]))
        return [s_truck_new, s_drone_new]   

#### select a neighborhood #########
    def neighborhoods(self, s_truck, s_drone, operator):
        if operator == 't2d':
            [s_truck_current, s_drone_current] = self.genetic_t2d(s_truck, s_drone)

        elif operator == 'd2t':
            [s_truck_current, s_drone_current] = self.genetic_d2t(s_truck, s_drone)

        elif operator == 'swap':
            [s_truck_current, s_drone_current] = self.genetic_swap(s_truck, s_drone)

        elif operator == 'relocation':
            [s_truck_current, s_drone_current] = self.n_relocation_truck_route(s_truck, s_drone)

        elif operator == '2opt':
            [s_truck_current, s_drone_current] = self.n_2opt(s_truck, s_drone)

        elif operator == 'id_greedy':
            [s_truck_current, s_drone_current] = self.n_idgreedy(s_truck, s_drone)

        elif operator == 'id_random':
            [s_truck_current, s_drone_current] = self.n_idrandom(s_truck, s_drone)

        else:
            print('wrong operations')
            s_truck_current = None
            s_drone_current = None
            s               = None
        
        s = self.evaluate(s_truck_current, s_drone_current)

        return [s_truck_current, s_drone_current, s]



            
                    








            





        





